
from ..dataset.polygon import *
from ..basic.basic_sequence import SeqAccumulator
from ..basic.basic_objects import BasObj
from ..import_basic_utils import *
import pickle
import numpy as np
import cv2

class Cifar10Maker(SeqAccumulator):

    def acc_sum(self):
        super().acc_sum()
        out_dir = os.path.join(self.args.root_dir, self.args.traindb_dir)
        if not os.path.exists(out_dir): os.makedirs(out_dir)
        # train
        data = np.vstack(self.acc['train_array'])
        batch_label = 'training batch'
        data_dict = {'filenames': self.acc['train_filenames'],
                     'data': data,
                     'labels': self.acc['train_labels'],
                     'batch_label': batch_label}
        with open(os.path.join(out_dir, 'data_batch'), "wb") as f:
            pickle.dump(data_dict, f)
        # test
        data = np.vstack(self.acc['test_array'])
        batch_label = 'testing batch'
        data_dict = {'filenames': self.acc['test_filenames'],
                     'data': data,
                     'labels': self.acc['test_labels'],
                     'batch_label': batch_label}
        with open(os.path.join(out_dir, 'test_batch'), "wb") as f:
            pickle.dump(data_dict, f)
        # meta
        num_vis = self.args.input_shape[0]*self.args.input_shape[1]*self.args.input_shape[2]  # 3*32*32
        label_names = self.args.classes
        num_cases_per_batch = self.acc['total']
        meta_dict = {'num_vis': num_vis,
                     'label_names': label_names,
                     'num_cases_per_batch': num_cases_per_batch}
        with open(os.path.join(out_dir, 'batches.meta'), "wb") as f:
            pickle.dump(meta_dict, f)


    def to_train(self, count):
        dur = self.args.train_val_div[0] + self.args.train_val_div[1]
        return (count % dur) < self.args.train_val_div[0]

    def init_acc(self):
        super().init_acc()
        self.num_classes = len(self.args.classes)
        self.acc.update({'cls_inds': np.zeros((self.num_classes, ), np.uint32)})
        self.list_to_write.append('cls_inds')

        self.acc.update({'train_array':[], 'test_array':[],
                'train_filenames':[], 'test_filenames':[],
                'train_labels':[], 'test_labels': []})


    def process_single(self, rgbd=None, data=None):

        if data is None:
            if rgbd is None: rgbd=self.rgbd
            rgbd = rgbd.resize(self.args.input_shape[:2])
            if self.args.show_steps: rgbd.show(title='rgbd_resize',args=self.args)

            data = rgbd.array(get_rgb=self.args.get_rgb, get_depth=self.args.get_depth, depth2norm=self.args.depth2norm)
        if self.args.show_steps: cv2.imshow('depth', data[:,:,3:])
        data_shape = data.shape
        if len(data_shape)<3: data = np.expand_dims(data, axis=2)
        h, w, num_ch = data.shape
        data_1D_org = [data[:,:,ch].reshape((1, h * w)) for ch in range(num_ch)]

        org_color_order = list(range(num_ch))
        color_orders = [org_color_order,]

        if hasattr(self.args, 'aug_color_orders') and self.args.get_rgb:
            for color_order in self.args.aug_color_orders:
                color_orders.append(list(color_order) + org_color_order[3:])

        for color_order in color_orders:
            data_1D = [data_1D_org[ch] for ch in color_order]
            data_1D = np.hstack(data_1D)

            data_reorder = data[:,:, color_order]

            # for ch in range(data.shape[2]):
            #     ch_1D = data[:,:,ch].reshape((1, h * w))
            #     if data_1D is None: data_1D = ch_1D
            #     else: data_1D = np.concatenate((data_1D, ch_1D), axis=1)


            if  hasattr(self,'cls_ind'): cls_ind = self.cls_ind
            else: cls_ind = -1

            self.acc['cls_inds'][cls_ind] += 1

            # update
            if self.to_train(self.acc['total']):
                self.acc['train_array'].append(data_1D)
                self.acc['train_filenames'].append(self.filename)
                self.acc['train_labels'].append(cls_ind)

                num_train = self.acc['num_train']
                self.acc['mean'] = num_train/(num_train+1)*self.acc['mean'] +\
                              1/(num_train+1)*np.mean(data_reorder.astype('float')/255, axis=(0, 1))
                self.acc['std'] = num_train / (num_train + 1) * self.acc['std'] + \
                              1 / (num_train + 1) * np.std(data_reorder.astype('float')/255, axis=(0, 1))
                self.acc['num_train'] += 1
            else:
                self.acc['test_array'].append(data_1D)
                self.acc['test_filenames'].append(self.filename)
                self.acc['test_labels'].append(cls_ind)

                self.acc['num_test'] += 1

            self.acc['total'] = self.acc['num_train'] + self.acc['num_test']

        if self.args.show_steps:
            if cv2.waitKey() ==27: exit()





class CifarModifier(BasObj):
    def get_array(self, data_path):
        f= open(data_path, 'rb')
        entry = pickle.load(f)
        f.close()
        return entry

    def load_params(self, args):
        super().load_params(args=args)
        self.train_filenames, self.test_filenames=None, None
        self.train_array, self.test_array = None, None
        self.train_labels, self.test_labels = None, None
        self.total = None
    def save_db(self, out_dir=None):

        if out_dir is None: save_db_dir = os.path.join(self.args.combine_dir, 'cifar10')
        else: save_db_dir = os.path.join(out_dir, 'cifar10')

        if not os.path.exists(save_db_dir): os.makedirs(save_db_dir)

        # train
        batch_label = 'training batch'
        data_dict = {'filenames': self.train_filenames,
                     'data': self.train_array,
                     'labels': self.train_labels,
                     'batch_label': batch_label}
        with open(os.path.join(save_db_dir,'data_batch'), "wb") as f:
            pickle.dump(data_dict, f, protocol=4)
        # test
        batch_label = 'testing batch'
        data_dict = {'filenames': self.test_filenames,
                     'data': self.test_array,
                     'labels': self.test_labels,
                     'batch_label': batch_label}
        with open(os.path.join(save_db_dir, 'test_batch'), "wb") as f:
            pickle.dump(data_dict, f, protocol=4)
        # meta
        num_vis = self.args.input_shape[0] * self.args.input_shape[1] * self.args.input_shape[2]  # 3*32*32
        label_names = self.args.classes
        num_cases_per_batch = self.total
        meta_dict = {'num_vis': num_vis,
                     'label_names': label_names,
                     'num_cases_per_batch': num_cases_per_batch}
        with open(os.path.join(save_db_dir, 'batches.meta'), "wb") as f:
            pickle.dump(meta_dict, f)


        # about
        about_db = open(os.path.join(save_db_dir, 'about_this_db'), 'w')
        list_to_write = ['num_train', 'num_test', 'total', 'mean', 'std', 'cls_inds']
        for name in list_to_write:
            value = self.__getattribute__(name)
            about_db.write('%s:\t%s\n' % (str(name), str(value)))
            np.save(os.path.join(save_db_dir, name + '.npy'), value)
        about_db.close()


class CombineCifar10Data(CifarModifier):
    def run(self):
        self.num_train, self.num_test, self.total = 0, 0, 0
        self.mean, self.std = 0,0
        self.cls_inds = 0
        self.train_array, self.test_array = None, None
        self.train_filenames, self.test_filenames = [], []
        self.train_labels, self.test_labels = [], []
        for db_path in self.args.db_paths:
            num_train = int(np.load(os.path.join(db_path, 'cifar10','num_train.npy')))
            num_test = int(np.load(os.path.join(db_path, 'cifar10', 'num_test.npy')))
            total = int(np.load(os.path.join(db_path, 'cifar10', 'total.npy')))
            mean = np.load(os.path.join(db_path, 'cifar10', 'mean.npy'))
            std = np.load(os.path.join(db_path, 'cifar10', 'std.npy'))
            cls_inds = np.load(os.path.join(db_path, 'cifar10', 'cls_inds.npy'))

            self.num_train += num_train
            self.num_test += num_test
            self.total += total
            self.mean += mean*total
            self.std += std*total
            self.cls_inds += cls_inds

            train_entry = self.get_array(os.path.join(db_path, 'cifar10', 'data_batch'))
            test_entry = self.get_array(os.path.join(db_path, 'cifar10', 'test_batch'))

            if self.train_array is None: self.train_array = train_entry['data']
            else: self.train_array = np.concatenate((self.train_array, train_entry['data']), axis=0)
            if self.test_array is None: self.test_array = test_entry['data']
            else: self.test_array = np.concatenate((self.test_array, test_entry['data']), axis=0)

            self.train_filenames += train_entry['filenames']
            self.test_filenames += test_entry['filenames']

            self.train_labels += train_entry['labels']
            self.test_labels += test_entry['labels']

        self.mean /= self.total
        self.std /= self.total

        self.save_db()

class CifarDepth2Norm(CifarModifier):
    def run(self, db_path=None):
        if db_path is None: db_path  = self.args.combine_dir

        self.num_train = int(np.load(os.path.join(db_path, 'cifar10', 'num_train.npy')))
        self.num_test = int(np.load(os.path.join(db_path, 'cifar10', 'num_test.npy')))
        self.total = int(np.load(os.path.join(db_path, 'cifar10', 'total.npy')))
        mean = np.load(os.path.join(db_path, 'cifar10', 'mean.npy'))
        std = np.load(os.path.join(db_path, 'cifar10', 'std.npy'))
        self.cls_inds = np.load(os.path.join(db_path, 'cifar10', 'cls_inds.npy'))

        print('loading data...')
        train_entry = self.get_array(os.path.join(db_path, 'cifar10', 'data_batch'))
        test_entry = self.get_array(os.path.join(db_path, 'cifar10', 'test_batch'))

        train_array = train_entry['data']
        test_array = test_entry['data']

        self.train_filenames = train_entry['filenames']
        self.test_filenames = test_entry['filenames']

        self.train_labels = train_entry['labels']
        self.test_labels = test_entry['labels']

        print('computing ...')
        h,w,ch = self.args.input_shape

        depth_train_array = train_array[:,-h*w:]
        depth_test_array = test_array[:,-h*w:]

        norm_train_array = [self.depth2norm1D(depth_train_array[i, :].reshape((h,w)))
                             for i in range(depth_train_array.shape[0])]
        norm_train_array = np.vstack(norm_train_array)

        norm_test_array = [self.depth2norm1D(depth_test_array[i, :].reshape((h, w)))
                             for i in range(depth_test_array.shape[0])]
        norm_test_array = np.vstack(norm_test_array)

        self.train_array = np.concatenate((train_array[:,-h*w], norm_train_array), axis=1)
        self.test_array = np.concatenate((train_array[:,-h * w], norm_test_array), axis=1)

        self.mean = mean + tuple(np.mean(norm_train_array[:, i:(i+1)*h*w]/255.) for i in range(3))
        self.std = std + tuple(np.std(norm_test_array[:, i:(i + 1) * h * w]/255.) for i in range(3))

        print('saving ...')
        self.save_db(out_dir=os.path.join(self.args.combine_dir + '_norm_vector'))


    def depth2norm1D(self, depth):
        norm_map = ArrayUtils().get_mat_normal_map_U8(depth)

        if self.args.show_steps:
            cv2.imshow('norm_map', norm_map)
            cv2.waitKey()
        h,w = self.args.input_shape[:2]
        return np.hstack([norm_map[:,:,i].reshape(1,h*w) for i in range(3)])




















from libs.utils.data_processing import *
from libs.policy.policy import *
from libs.policy.sequence_show_predict import SeqPredictor
from libs.policy.policy import GenDetector
from libs.dataset.polygon import *
from libs.dataset.grip import GripDBer


def detect_sequence_local(cfg_path):
    drawer = OnmousePolygon(cfg_path=cfg_path)
    SeqPredictor(cfg_path=cfg_path, detector=GenDetector(cfg_path=cfg_path, detector=drawer.get_polygon,
                                                     show_ret_func=show_polygon).predict_show, detector_reload=drawer.load_params).run()
    # drawer = DrawPolyon(cfg_path=cfg_path)
    # SequencePredictor(cfg_path, detector=GenDetector(cfg_path=cfg_path, detector=drawer.get_ann,
    #                                                  show_ret_func=show_polygon).predict_show, detector_reload=drawer.load_params).run()

    # drawer = GripDBer(cfg_path=cfg_path)
    # SeqPredictor(cfg_path=cfg_path, detector=GenDetector(cfg_path=cfg_path, detector=drawer.poly2maskAug).predict_show,
    #                   detector_reload=drawer.load_params).run()
if __name__ == '__main__':
    detect_sequence_local('configs/grasp_detection/grip_evaluator.cfg')





from __future__ import division
""" 
Creates a ResNeXt Model as defined in:
Xie, S., Girshick, R., Dollar, P., Tu, Z., & He, K. (2016). 
Aggregated residual transformations for deep neural networks. 
arXiv preprint arXiv:1611.05431.
import from https://github.com/facebookresearch/ResNeXt/blob/master/models/resnext.lua
"""
import math
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init
import torch

__all__ = ['resnext50', 'resnext101', 'resnext152']

class Bottleneck(nn.Module):
    """
    RexNeXt bottleneck type C
    """
    expansion = 4

    def __init__(self, inplanes, planes, baseWidth, cardinality, stride=1, downsample=None):
        """ Constructor
        Args:
            inplanes: input channel dimensionality
            planes: output channel dimensionality
            baseWidth: base width.
            cardinality: num of convolution groups.
            stride: conv stride. Replaces pooling layer.
        """
        super(Bottleneck, self).__init__()

        D = int(math.floor(planes * (baseWidth / 64)))
        C = cardinality

        self.conv1 = nn.Conv2d(inplanes, D*C, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn1 = nn.BatchNorm2d(D*C)
        self.conv2 = nn.Conv2d(D*C, D*C, kernel_size=3, stride=stride, padding=1, groups=C, bias=False)
        self.bn2 = nn.BatchNorm2d(D*C)
        self.conv3 = nn.Conv2d(D*C, planes * 4, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)

        self.downsample = downsample

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNeXt(nn.Module):
    """
    ResNext optimized for the ImageNet dataset, as specified in
    https://arxiv.org/pdf/1611.05431.pdf
    """
    def __init__(self, baseWidth, cardinality, layers, num_classes):
        """ Constructor
        Args:
            baseWidth: baseWidth for ResNeXt.
            cardinality: number of convolution groups.
            layers: config of layers, e.g., [3, 4, 6, 3]
            num_classes: number of classes
        """
        super(ResNeXt, self).__init__()
        block = Bottleneck

        self.cardinality = cardinality
        self.baseWidth = baseWidth
        self.num_classes = num_classes
        self.inplanes = 64
        self.output_size = 64

        self.conv1 = nn.Conv2d(3, 64, 7, 2, 3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], 2)
        self.layer3 = self._make_layer(block, 256, layers[2], 2)
        self.layer4 = self._make_layer(block, 512, layers[3], 2)
        self.avgpool = nn.AvgPool2d(7)      
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        """ Stack n bottleneck modules where n is inferred from the depth of the network.
        Args:
            block: block type used to construct ResNext
            planes: number of output channels (need to multiply by block.expansion)
            blocks: number of blocks to be built
            stride: factor to reduce the spatial dimensionality in the first bottleneck of the block.
        Returns: a Module consisting of n sequential bottlenecks.
        """
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, self.baseWidth, self.cardinality, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, self.baseWidth, self.cardinality))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool1(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


def resnext50(baseWidth, cardinality):
    """
    Construct ResNeXt-50.
    """
    model = ResNeXt(baseWidth, cardinality, [3, 4, 6, 3], 1000)
    return model


def resnext101(baseWidth, cardinality):
    """
    Construct ResNeXt-101.
    """
    model = ResNeXt(baseWidth, cardinality, [3, 4, 23, 3], 1000)
    return model


def resnext152(baseWidth, cardinality):
    """
    Construct ResNeXt-152.
    """
    model = ResNeXt(baseWidth, cardinality, [3, 8, 36, 3], 1000)
    return model



import torch.nn as nn
import torch.nn.functional as F
from ..basic.models.resnet import ResNet
from ..basic.models.mobilenet import MobileNetV2


__all__ = ['Net','SimpleNet', 'ResNet', 'MobileNetV2']


def num_flat_features(x):
    size = x.size()[1:]  # all dimensions except the batch dimension
    num_features = 1
    for s in size:
        num_features *= s
    return num_features

class SimpleNet(nn.Module):

    def __init__(self, input_shape=(32, 32, 3), num_classes=10):
        super(SimpleNet, self).__init__()
        h, w, ch = input_shape
        # kernels
        self.conv1 = nn.Conv2d(ch, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        # an affine operation: y = Wx + b
        # self.fc1 = nn.Linear(16 * int((h-4)/4) * int((w-8)/4), 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, num_classes)             # 5: number of classes  (without background)
        # self.dout = nn.Dropout(0.5)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, num_flat_features(x))
        # x = F.relu(self.fc1(x))
        x = F.relu(nn.Linear(num_flat_features(x), 120)(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x





def Net(model_name, **kwargs):

    if model_name not in __all__:
        print('No model named %s' %model_name)
        return None

    return eval(model_name)(**kwargs)



"""
Copyright (c) 2019-present NAVER Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import torch.nn as nn

from .transformation import TPS_SpatialTransformerNetwork
from .feature_extraction import VGG_FeatureExtractor, RCNN_FeatureExtractor, ResNet_FeatureExtractor
from .sequence_modeling import BidirectionalLSTM
from .prediction import Attention


class Model(nn.Module):

    def __init__(self, args):
        super(Model, self).__init__()
        self.args = args
        self.stages = {'Trans': args['transformation'], 'Feat': args['featureextraction'],
                       'Seq': args['sequencemodeling'], 'Pred': args['prediction']}

        """ Transformation """
        if args['transformation'] == 'TPS':
            self.Transformation = TPS_SpatialTransformerNetwork(
                F=args['num_fiducial'], I_size=(args['imgh'], args['imgw']),
                I_r_size=(args['imgh'], args['imgw']), I_channel_num=args['input_channel'])
        else:
            print('No Transformation module specified')

        """ FeatureExtraction """
        if args['featureextraction'] == 'VGG':
            self.FeatureExtraction = VGG_FeatureExtractor(args['input_channel'], args['output_channel'])
        elif args['featureextraction'] == 'RCNN':
            self.FeatureExtraction = RCNN_FeatureExtractor(args['input_channel'], args['output_channel'])
        elif args['featureextraction'] == 'ResNet':
            self.FeatureExtraction = ResNet_FeatureExtractor(args['input_channel'], args['output_channel'])
        else:
            raise Exception('No FeatureExtraction module specified')
        self.FeatureExtraction_output = args['output_channel']  # int(imgH/16-1) * 512
        self.AdaptiveAvgPool = nn.AdaptiveAvgPool2d((None, 1))  # Transform final (imgH/16-1) -> 1

        """ Sequence modeling"""
        if args['sequencemodeling'] == 'BiLSTM':
            self.SequenceModeling = nn.Sequential(
                BidirectionalLSTM(self.FeatureExtraction_output, args['hidden_size'], args['hidden_size']),
                BidirectionalLSTM(args['hidden_size'], args['hidden_size'], args['hidden_size']))
            self.SequenceModeling_output = args['hidden_size']
        else:
            print('No SequenceModeling module specified')
            self.SequenceModeling_output = self.FeatureExtraction_output

        """ Prediction """
        if args['prediction'] == 'CTC':
            self.Prediction = nn.Linear(self.SequenceModeling_output, args['num_class'])
        elif args['prediction'] == 'Attn':
            self.Prediction = Attention(self.SequenceModeling_output, args['hidden_size'], args['num_class'])
        else:
            raise Exception('Prediction is neither CTC or Attn')

    def forward(self, input, text, is_train=True):
        """ Transformation stage """
        if not self.stages['Trans'] == "None":
            input = self.Transformation(input)

        """ Feature extraction stage """
        visual_feature = self.FeatureExtraction(input)
        visual_feature = self.AdaptiveAvgPool(visual_feature.permute(0, 3, 1, 2))  # [b, c, h, w] -> [b, w, c, h]
        visual_feature = visual_feature.squeeze(3)

        """ Sequence modeling stage """
        if self.stages['Seq'] == 'BiLSTM':
            contextual_feature = self.SequenceModeling(visual_feature)
        else:
            contextual_feature = visual_feature  # for convenience. this is NOT contextually modeled by BiLSTM

        """ Prediction stage """
        if self.stages['Pred'] == 'CTC':
            prediction = self.Prediction(contextual_feature.contiguous())
        else:
            prediction = self.Prediction(contextual_feature.contiguous(), text, is_train, batch_max_length=self.args['batch_max_length'])

        return prediction



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='ttdet', 
    version='1.0.4',
    author="Trung M. Bui",
    author_email="bmtrungvp@gmail.com",
    description="A detection utilities for computer vision",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mtbui2010",
    packages= setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    #entry_points = {
    #      'console_scripts': ['ttcv_sample=ttcv.samples.select_sample:run'],
    # }
    license='MIT', 
    keywords = ['AI','VISION', 'DETECTION'],
    install_requires=[            
          'numpy',
        'opencv-python',
         'scipy',
         'matplotlib',
   	'ttcv',
        'torch==1.4.0',
        'torchvision==0.5.0',
	'progress',
     ],
)	



from __future__ import absolute_import

'''Resnet for cifar dataset.
Ported form
https://github.com/facebook/fb.resnet.torch
and
https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
(c) YANG, Wei
'''
import torch.nn as nn
import math


__all__ = ['resnet']

def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, depth, input_shape=(32,32,3),num_classes=1000, block_name='BasicBlock'):
        super(ResNet, self).__init__()
        # Model type specifies number of layers for CIFAR-10 model
        if block_name.lower() == 'basicblock':
            assert (depth - 2) % 6 == 0, 'When use basicblock, depth should be 6n+2, e.g. 20, 32, 44, 56, 110, 1202'
            n = (depth - 2) // 6
            block = BasicBlock
        elif block_name.lower() == 'bottleneck':
            assert (depth - 2) % 9 == 0, 'When use bottleneck, depth should be 9n+2, e.g. 20, 29, 47, 56, 110, 1199'
            n = (depth - 2) // 9
            block = Bottleneck
        else:
            raise ValueError('block_name shoule be Basicblock or Bottleneck')

        h,w, ch = input_shape
        self.num_classes = num_classes
        self.inplanes = 16
        self.conv1 = nn.Conv2d(ch, 16, kernel_size=3, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 16, n)
        self.layer2 = self._make_layer(block, 32, n, stride=2)
        self.layer3 = self._make_layer(block, 64, n, stride=2)
        self.avgpool = nn.AvgPool2d(8)
        self.fc = nn.Linear(int(h * w/16) * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)    # 32x32

        x = self.layer1(x)  # 32x32
        x = self.layer2(x)  # 16x16
        x = self.layer3(x)  # 8x8

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        # x = nn.Linear(num_flat_features(x), self.num_classes)(x)

        return x

def num_flat_features(x):
    size = x.size()[1:]  # all dimensions except the batch dimension
    num_features = 1
    for s in size:
        num_features *= s
    return num_features



def resnet(**kwargs):
    """
    Constructs a ResNet model.
    """
    return ResNet(**kwargs)



from mmdet.apis import init_detector, inference_detector, show_result
import mmcv
import numpy as np
import pycocotools.mask as maskUtils
from time import time
import configparser
import cv2
from libs.utils.proc_utils import get_color, correct_workspace
from libs.basic.basic_objects import BasDetectObj

class MaskRCNNDetector(BasDetectObj):

    def get_model(self):
        self.model = init_detector(self.args['config_file'], self.args['checkpoint_file'], device='cuda:0')

    def predict_array(self, inputs):
        t = time()
        print('---------------------------->>>>>>>>>> Mask RCNN detection')

        self.crop_rect = correct_workspace(inputs['im'].shape[:2], self.args['crop_rect'], self.args['crop_im'])
        # self.crop_rect = self.args['crop_rect']

        crop_im =inputs['im'][self.crop_rect[0]:self.crop_rect[2], self.crop_rect[1]:self.crop_rect[3], :]
        boxes_list, masks_list = inference_detector(self.model, inputs['im'][self.crop_rect[0]:self.crop_rect[2], self.crop_rect[1]:self.crop_rect[3], :])

        num_boxes = 0
        for i, cls in enumerate(self.args['classes']):
            boxes = boxes_list[i]
            masks = masks_list[i]
            if len(boxes) == 0: continue
            num_boxes += len(boxes)
            new_masks = []
            # new_boxes = []
            for j in range(len(boxes)):
                mask = 255*maskUtils.decode(masks[j]).astype(np.uint8)
                left, top, right, bottom = boxes[j,:4].astype('int')

                new_masks.append(mask[top:bottom, left:right])
                boxes[j, :4] = left + self.crop_rect[1], top + self.crop_rect[0], right + self.crop_rect[1], bottom + self.crop_rect[0]

            boxes_list[i] = boxes
            masks_list[i] = new_masks

        print('Detected %d objects in %.3f s' % (num_boxes, (time() - t)))
        return {'boxes':boxes_list, 'masks':masks_list}

    def show_detect_results(self, im, detect_rets):
        boxes_list = detect_rets['boxes']
        masks_list = detect_rets['masks']
        out = np.copy(im)
        obj_ind = 0
        for j, cls in  enumerate(self.args['classes']):
            boxes = boxes_list[j]
            num_boxes = len(boxes)
            if num_boxes == 0 or cls == '__background__': continue
            masks = masks_list[j]

            for box, mask in zip(boxes, masks):
                color = get_color(obj_ind)

                left, top, right, bottom, prob = box
                left, top, right, bottom = int(left), int(top), int(right), int(bottom)

                Y, X = np.where(mask==255)
                Y += top
                X += left
                loc = (Y,X)

                out[loc] = 0.8 * out[loc] + tuple(0.2 * np.array(color))
                cv2.rectangle(out, (left, top), (right, bottom), color, 2)
                cv2.putText(out, '%s: %.2f' %(cls, prob), (left, top-3), cv2.FONT_HERSHEY_SIMPLEX, 0.5, color)

                obj_ind += 1
        return out










from ttcv.basic.basic_objects import BasObj
from ..basic.basic_classification_models import Net
import torchvision.transforms as transforms
import torch, torchvision
import torch.nn as nn
import torch.optim as optim
import os
import numpy as np
from ..dataset.datasets import Dataset, get_mean_std
import random
from ..cifar_classfication import cifar_classification_models as models
import torch.backends.cudnn as cudnn
import shutil
from time import time

from ..cifar_classfication.classification_utils import Logger, AverageMeter, accuracy, mkdir_p, savefig
from ..cifar_classfication.classification_utils.progress.progress.bar import Bar
from PIL import Image
from torch.autograd import Variable
import torch.nn.functional as F
from ttcv.import_basic_utils import *



class BasCifarClassfier(BasObj):
    def __init__(self,args=None, cfg_path=None, name='unnamed'):
        super().__init__(args=args, cfg_path=cfg_path, name=name)
        self.get_model()

    def load_params(self, args):
        super().load_params(args=args)
        self.state = args.todict()
        self.get_data()

    def predict_array(self,im):
        # t = time()
        # img = Image.fromarray(im)
        img = self.transform_test(im)
        if self.use_cuda: img = img.cuda()
        img = img.view(1, *img.size())
        img = Variable(img)
        predicted = self.model(img)
        probs = F.softmax(predicted, dim=1)
        if self.use_cuda: probs = probs.cpu()
        # print('classification elapsed: %.3fs' %(time()-t))
        return probs.detach().numpy().flatten()

    def predict_rgbd(self, rgbd):
        im = rgbd.resize(self.args.input_shape[:2]).array(get_rgb=self.args.get_rgb,
                                                          get_depth=self.args.get_depth,
                                                          depth2norm=self.args.depth2norm)
        return self.predict_array(im)

    def predict_arrays_m(self, ims, print_info=True):
        timer = Timer()
        img_concat = self.arrays2tensors(ims=ims)
        timer.pin_time(label='concat')
        predicted = self.model(img_concat)
        probs = F.softmax(predicted, dim=1)
        if self.use_cuda: probs = probs.cpu()
        probs = probs.detach().numpy()
        timer.pin_time(label='predict')
        if print_info: print(timer.pin_times_str())
        return probs

    def arrays2tensors(self, ims):
        img_concat = []
        for im in ims:
            img = self.transform_test(im)
            if self.use_cuda: img = img.cuda()
            img_concat.append(img)
        img_concat = torch.stack(img_concat)
        # img_concat= torch.autograd.Variable(img_concat, volatile=True)
        return img_concat

    # def predict_arrays(self, ims, print_info=True):
    #     timer = Timer()
    #     num_im = len(ims)
    #     if num_im <= self.args.test_batch: return self.predict_arrays_m(ims)
    #
    #     part = list(np.arange(0, num_im, self.args.test_batch))
    #     if part[-1] != num_im: part.append(num_im)
    #
    #     probs = []
    #     for i in range(len(part)-1):
    #         select_ims = ims[part[i]: part[i+1]]
    #         partial_probs = self.predict_arrays_m(select_ims, print_info=False)
    #         probs.append(partial_probs)
    #     probs = np.vstack(probs)
    #
    #     if print_info:
    #         print('test batch size: %d' %self.args.test_batch)
    #         print(timer.pin_times_str())
    #
    #
    #     return probs

    def predict_arrays(self, ims, print_info=True):
        num_im = len(ims)
        if num_im<self.args.test_batch: return self.predict_arrays_m(ims=ims, print_info=print_info)
        timer1 = Timer()
        testset = Dataset(name=self.args.dber , train=False, data=ims, im_shape=self.args.input_shape, transform=self.transform_test, indexing=True)
        testloader = torch.utils.data.DataLoader(testset, batch_size=self.args.test_batch,
                                                      shuffle=True, num_workers=self.args.num_workers)
        timer1.pin_time('load_data')

        num_batch = int(np.ceil(num_im/self.args.test_batch))
        probs = np.zeros((num_im, self.num_classes))

        # probs_tensor = torch.zeros((num_im, self.num_classes), dtype=torch.float,device=self.device)
        for batch_idx, (inputs, targets, indexes) in enumerate(testloader):
            timer = Timer()
            if self.use_cuda:
                inputs = inputs.cuda()
            # inputs = torch.autograd.Variable(inputs, volatile=True)
            outputs = self.model(inputs)
            outputs = F.softmax(outputs, dim=1)
            timer.pin_time('score')

            if self.use_cuda: outputs = outputs.cpu()
            outputs = outputs.detach().numpy()
            timer.pin_time('cvt_cpu')

            probs[(indexes,)] = outputs
            timer.pin_time('save_array')
            # print(timer.pin_times_str())

        timer1.pin_time('scoring_all')
        if print_info:
            print('batch_szie:%d-num_batch:%d-msPerBatch:%0.1f' %(self.args.test_batch, num_batch,
                                                                  1000*timer1.get_pinned_time('scoring_all')/num_batch))
            print(timer1.pin_times_str())

        return probs


    def predict_rgbds(self, rgbds):
        ims = []
        for rgbd in rgbds:
            ims.append(rgbd.resize(self.args.input_shape[:2]).array(get_rgb=self.args.get_rgb,
                                                                    get_depth=self.args.get_depth,
                                                                    depth2norm=self.args.depth2norm))
        probs = self.predict_arrays(ims)
        # print('classification elapsed: %.3fs' % (time() - t))
        return probs


    def trainval(self):
        # Train and val
        for epoch in range(self.start_epoch, self.args.epochs):
            self.adjust_learning_rate(epoch)

            print('\nEpoch: [%d | %d] LR: %f' % (epoch + 1, self.args.epochs, self.state['lr']))

            train_loss, train_acc = self.train()
            test_loss, test_acc = self.test()

            # append logger file
            self.logger.append([self.state['lr'], train_loss, test_loss, train_acc, test_acc])

            # save model
            is_best = test_acc > self.best_acc
            self.best_acc = max(test_acc, self.best_acc)
            self.save_checkpoint({
                'epoch': epoch + 1,
                'state_dict': self.model.state_dict(),
                'acc': test_acc,
                'best_acc': self.best_acc,
                'optimizer': self.optimizer.state_dict(),
            }, is_best)

        self.logger.close()
        self.logger.plot()
        savefig(os.path.join(self.args.checkpoint_dir, 'log.eps'))

        print('Best acc:')
        print(self.best_acc)


    def train(self):
        # switch to train mode
        self.model.train()

        batch_time = AverageMeter()
        data_time = AverageMeter()
        losses = AverageMeter()
        top1 = AverageMeter()
        top5 = AverageMeter()
        end = time()

        bar = Bar('Processing', max=len(self.trainloader))
        for batch_idx, (inputs, targets) in enumerate(self.trainloader):
            # measure data loading time
            data_time.update(time() - end)

            if self.use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda(async=True)
            inputs, targets = torch.autograd.Variable(inputs), torch.autograd.Variable(targets)

            # compute output
            outputs = self.model(inputs)
            loss = self.criterion(outputs, targets)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(outputs.data, targets.data, topk=(1,2))
            losses.update(loss.item(), inputs.size(0))
            top1.update(prec1.item(), inputs.size(0))
            top5.update(prec5.item(), inputs.size(0))

            # compute gradient and do SGD step
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # measure elapsed time
            batch_time.update(time() - end)
            end = time()

            # plot progress
            bar.suffix = '({batch}/{size}) Data: {data:.3f}s | Batch: {bt:.3f}s | Total: {total:} | ETA: {eta:} | Loss: {loss:.4f} | top1: {top1: .4f} | top5: {top5: .4f}'.format(
                batch=batch_idx + 1,
                size=len(self.trainloader),
                data=data_time.avg,
                bt=batch_time.avg,
                total=bar.elapsed_td,
                eta=bar.eta_td,
                loss=losses.avg,
                top1=top1.avg,
                top5=top5.avg,
            )
            bar.next()
        bar.finish()
        return (losses.avg, top1.avg)

    def test(self):
        batch_time = AverageMeter()
        data_time = AverageMeter()
        losses = AverageMeter()
        top1 = AverageMeter()
        top5 = AverageMeter()

        # switch to evaluate mode
        self.model.eval()

        end = time()
        bar = Bar('Processing', max=len(self.testloader))
        for batch_idx, (inputs, targets) in enumerate(self.testloader):
            # measure data loading time
            data_time.update(time() - end)

            if self.use_cuda:
                inputs, targets = inputs.cuda(), targets.cuda()
            inputs, targets = torch.autograd.Variable(inputs, volatile=True), torch.autograd.Variable(targets)

            # compute output
            outputs = self.model(inputs)
            loss = self.criterion(outputs, targets)

            # measure accuracy and record loss
            prec1, prec5 = accuracy(outputs.data, targets.data, topk=(1, 2))
            losses.update(loss.item(), inputs.size(0))
            top1.update(prec1.item(), inputs.size(0))
            top5.update(prec5.item(), inputs.size(0))

            # measure elapsed time
            batch_time.update(time() - end)
            end = time()

            # plot progress
            bar.suffix = '({batch}/{size}) Data: {data:.3f}s | Batch: {bt:.3f}s | Total: {total:} | ETA: {eta:} | Loss: {loss:.4f} | top1: {top1: .4f} | top5: {top5: .4f}'.format(
                batch=batch_idx + 1,
                size=len(self.testloader),
                data=data_time.avg,
                bt=batch_time.avg,
                total=bar.elapsed_td,
                eta=bar.eta_td,
                loss=losses.avg,
                top1=top1.avg,
                top5=top5.avg,
            )
            bar.next()
        bar.finish()
        return (losses.avg, top1.avg)



    def get_data(self):
        # Transform
        db_mean, db_std = get_mean_std(self.args.dber)

        self.transform_train = transforms.Compose([
            # transforms.RandomCrop(32, padding=4),
            # transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(db_mean, db_std),
        ])
        self.transform_test = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(db_mean, db_std),
        ])
        if self.args.train_mode:
            # Data
            if not hasattr(self.args, 'combine_dir'): data_root = self.args.root_dir
            else: data_root = self.args.combine_dir
            testset = Dataset(name=self.args.dber, root=data_root, train=False, download=self.args.db_download,
                              transform=self.transform_test, im_shape=self.args.input_shape)
            self.testloader = torch.utils.data.DataLoader(testset, batch_size=self.args.test_batch,
                                                          shuffle=True, num_workers=self.args.num_workers)

            trainset = Dataset(name=self.args.dber, root=data_root, train=True, download=self.args.db_download,
                               transform=self.transform_train, im_shape=self.args.input_shape)
            self.trainloader = torch.utils.data.DataLoader(trainset, batch_size=self.args.train_batch,
                                                           shuffle=True, num_workers=self.args.num_workers)
            self.classes = trainset.classes
        else: self.classes = self.args.classes

        self.num_classes = len(self.classes)

    def get_model(self):
        self.use_cuda = torch.cuda.is_available()
        self.device = torch.device("cuda:0" if self.use_cuda else "cpu")

        # Model
        print("==> creating model '{}'".format(self.args.arch))
        if self.args.arch.startswith('resnext'):
            model = models.__dict__[self.args.arch](
                cardinality=self.args.cardinality,
                num_classes=self.num_classes,
                depth=self.args.model_depth,
                widen_factor=self.args.widen_factor,
                dropRate=self.args.drop,
            )
        elif self.args.arch.startswith('densenet'):
            model = models.__dict__[self.args.arch](
                num_classes=self.num_classes,
                depth=self.args.depth,
                growthRate=self.args.growthRate,
                compressionRate=self.args.compressionRate,
                dropRate=self.args.drop,
            )
        elif self.args.arch.startswith('wrn'):
            model = models.__dict__[self.args.arch](
                num_classes=self.num_classes,
                depth=self.args.model_depth,
                widen_factor=self.args.widen_factor,
                dropRate=self.args.drop,
            )
        elif self.args.arch.endswith('resnet'):
            model = models.__dict__[self.args.arch](
                num_classes=self.num_classes,
                depth=self.args.model_depth,
                input_shape=self.args.input_shape
                # block_name=self.args.block_name,
            )
        else:
            model = models.__dict__[self.args.arch](num_classes=self.num_classes)
        self.model = torch.nn.DataParallel(model).cuda()

        cudnn.benchmark = True
        self.title = '%s-%s-' % (self.args.dber, self.args.arch)
        self.checkpoint_prefix = os.path.join(self.args.checkpoint_dir, self.title)
        if self.args.train_mode:
            self.criterion = nn.CrossEntropyLoss()
            self.optimizer = optim.SGD(self.model.parameters(), lr=self.args.lr, momentum=self.args.momentum,
                                       weight_decay=self.args.weight_decay)
            self.resume()
        else: self.load_model()

    def load_model(self):
        checkpoint_path = self.checkpoint_prefix + 'model_best.pth'
        assert os.path.exists(checkpoint_path)
        checkpoint = torch.load(checkpoint_path)
        self.model.load_state_dict(checkpoint['state_dict'])
        self.model.eval()

    def resume(self):
        # save
        if not os.path.exists(self.args.checkpoint_dir): os.makedirs(self.args.checkpoint_dir)

        # Resume
        self.resume_checkpoint = self.checkpoint_prefix + 'model_best.pth'
        has_resume = os.path.exists(self.resume_checkpoint)

        if has_resume:
            # Load checkpoint.
            print('==> Resuming from checkpoint..')
            checkpoint = torch.load(self.resume_checkpoint)
            self.best_acc = checkpoint['best_acc']
            self.start_epoch = checkpoint['epoch']
            self.model.load_state_dict(checkpoint['state_dict'])
            self.optimizer.load_state_dict(checkpoint['optimizer'])
            self.logger = Logger(self.checkpoint_prefix + 'log.txt', title=self.title, resume=True)
        else:
            self.logger = Logger(self.checkpoint_prefix + 'log.txt', title=self.title)
            self.logger.set_names(['Learning Rate', 'Train Loss', 'Valid Loss', 'Train Acc.', 'Valid Acc.'])
            self.best_acc = 0
            self.start_epoch = 0

    def save_checkpoint(self, state, is_best):
        filepath = self.checkpoint_prefix + str(state['epoch']).rjust(4, '0') + '.pth'
        temp = self.checkpoint_prefix +  'temp.pth'
        torch.save(state, temp)
        if is_best:
            shutil.copyfile(temp, self.resume_checkpoint)
        if state['epoch'] % self.args.save_every == 0:
            shutil.copyfile(temp, filepath)

    def adjust_learning_rate(self, epoch):
        if epoch in self.args.schedule:
            self.state['lr'] *= self.args.gamma
            for param_group in self.optimizer.param_groups:
                param_group['lr'] = self.state['lr']


class RGBDLoader():

    def __init__(self, args, rgbds, transform):
        self.transform = transform
        self.ims = []
        for rgbd in rgbds:
            self.ims.append(rgbd.resize(args.input_shape[:2]).array(get_rgb=args.get_rgb,
                                                                    get_depth=args.get_depth))
    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: image.
        """
        img = self.ims[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)
        return img

    def __len__(self):
        return len(self.ims)










"""Useful utils
"""
from .misc import *
from .logger import *
from .visualize import *
from .eval import *

# progress bar
import os, sys
sys.path.append(os.path.join(os.path.dirname(__file__), "progress"))
from progress.bar import Bar as Bar


import cv2
import numpy as np
from matplotlib import pyplot as plt
from glob import  glob
import os

def matching(img_gray, template, threshold=0.5):
    height, width = template.shape[:2]
    res = cv2.matchTemplate(img_gray, template, cv2.TM_CCOEFF_NORMED)
    Y,X = np.where( res >= threshold)
    boxes, scores = [], []
    for y,x in zip(Y,X):
        boxes.append([int(x), int(y), int(x+width), int(y+height)])
        scores.append(float(res[(y,x)]))
    # print('{} {} boxes matched'.format('+'*10, len(boxes)))
    return boxes, scores

def matching_multi(img_gray, templates,threshold=0.5):
    boxes, scores = [], []
    for tmp_im in templates:
        bb, ss= matching(img_gray=img_gray, template=tmp_im, threshold=threshold)
        boxes += bb
        scores += ss
    return boxes, scores

def matching_dir(img_gray, template_dir,threshold=0.5):
    tmp_ims = [cv2.imread(path, cv2.IMREAD_GRAYSCALE)
               for path in glob(os.path.join(template_dir, '*'))]
    boxes, scores = matching_multi(img_gray, tmp_ims, threshold=threshold)
    # print('{} {} boxes matched'.format('+'*10, len(boxes)))
    return boxes, scores


if __name__=='__main__':
    img_rgb = cv2.imread('test_images/SourceIMG.jpeg')
    img_gray = cv2.cvtColor(img_rgb, cv2.COLOR_BGR2GRAY)
    template = cv2.imread('test_images/TemplateIMG.jpeg', 0)

    boxes = match(img_gray, template)
    for box in boxes:
        cv2.rectangle(img_rgb, (box[0], box[1]), (box[2], box[3]), (0,255,0), 2)

    cv2.imshow('im', img_rgb)
    cv2.waitKey()



from PIL import Image
import os
import os.path
import numpy as np
import sys
import torchvision.transforms as transforms
from ttcv.import_basic_utils import *

if sys.version_info[0] == 2:
    import cPickle as pickle
else:
    import pickle

from torchvision.datasets.vision import VisionDataset
from torchvision.datasets.utils import download_url, check_integrity


class SuctionCifar10(VisionDataset):

    base_folder = 'cifar10'
    url = None
    filename = None
    tgz_md5 = None
    train_list = [['data_batch', None]]

    test_list = [['test_batch', None]]
    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': None,
    }

    # db_mean = (0.37333256, 0.50367485, 0.6492233, 0.56081723)  # 20200511
    # db_std = (0.0414015, 0.03815473, 0.04328673, 0.01134201)

    # db_mean = (0.43503931, 0.44019999, 0.41939172, 0.55336628)  # 20200512
    # db_std = (0.06111116, 0.06243232, 0.06333934, 0.01970969)

    # db_mean = (0.39287165, 0.4465874,  0.40430869, 0.54449114)  # 20200512
    # db_std = (0.07679449, 0.07307834, 0.07486557, 0.02271312)

    # db_mean = (0.39583588, 0.4751535,  0.4489659,  0.39239821)  # 20200514
    # db_std = (0.09845068, 0.08866881, 0.08950291, 0.02083357)

    # db_mean = (0.61325816, 0.63284441, 0.62099887, 0.3794783)  # 20200521
    # db_std = (0.10611203, 0.10823587, 0.10509778, 0.02259126)

    # db_mean = (0.60783364, 0.62808136, 0.61624238, 0.38024026)  # 20200521_RB_best , (100, 60000)
    # db_std = (0.11982953, 0.11713887, 0.11431467, 0.02293329)

    # db_mean = (0.60783364, 0.62808136, 0.61624238)  # 20200521_v2 only rgb
    # db_std = (0.11982953, 0.11713887, 0.11431467)

    # db_mean = (0.41312494, 0.33027523, 0.262142,   0.6117158)  # 20200525_CH robot env
    # db_std = (0.08633106, 0.09178339, 0.10141662, 0.03975779)

    # db_mean = (0.51066111, 0.47836661, 0.43790377, 0.49762565)  # 20200525_RB_CH robot env
    # db_std = (0.0961087,  0.09922036, 0.1026221,  0.03070719)

    # db_mean = (0.60783364, 0.62808136, 0.61624238, 0.41772863)  # 20200527_RB_good
    # db_std = (0.11982953, 0.11713887, 0.11431467, 0.02475438)

    # db_mean = (0.6100014,  0.62563852, 0.60814321, 0.41632976)  # 20200527_RB_cmb_0519_0526
    # db_std = (0.11946112, 0.11505818, 0.11240116, 0.02360522)

    # db_mean = (0.50895616, 0.4740457,  0.4305609,  0.5415007)  # 20200527_RB_CH_cmb_0519_0522_0526
    # db_std = (0.10245739, 0.10311258, 0.10676344, 0.03361819)

    # db_mean = (0.49576933, 0.49576513, 0.49576391, 0.51792066)  # 20200604_RB_CH_cmb_0519_0522_0526 (fg, bound, bg)
    # db_std = (0.11185922, 0.11186486, 0.11187794, 0.03231334)

    # db_mean = (0.56726268, 0.56723482, 0.56719619, 0.47881223)  # 20200604_RB_CH_cmb_0519_0522_0526 (fg, bound)
    # db_std = (0.13903961, 0.13904731, 0.13908207, 0.03555515)

    db_mean = (0.55789759, 0.55786037, 0.55780825, 0.48949334)  # 20200608_RB_CH_cmb_0519_0522_0526_0604_0605 (fg, bound)
    db_std = (0.13491044, 0.13492254, 0.13494867, 0.03607968)

    # db_mean = (0.5639651,  0.56381155, 0.56385888, 0.18956837, 0.18703041, 0.88085926)  # 20200616_RB_CH_cmb_0519_0522_0526_0604_0605_norm_vec
    # db_std = (0.1361991,  0.13618944, 0.13621452, 0.19232603, 0.22332118, 0.13067493)



    def __init__(self, root=None, train=True,
                 transform=None, target_transform=None,
                 download=False, im_shape=(32,32,3), data=None, indexing=False):

        super(SuctionCifar10, self).__init__(root)
        self.transform = transform
        self.target_transform = target_transform

        self.indexing=indexing

        self.train = train  # training set or test set
        if download:
            self.download()

        if data is None:
            if not self._check_integrity():
                raise RuntimeError('Dataset not found or corrupted.' +
                                   ' You can use download=True to download it')

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data = []
        self.targets = []
        if data is not None: self.targets=[0]*len(data)
        timer = Timer()
        if data is None:
            # now load the picked numpy arrays
            for file_name, checksum in downloaded_list:
                file_path = os.path.join(self.root, self.base_folder, file_name)
                with open(file_path, 'rb') as f:
                    if sys.version_info[0] == 2:
                        entry = pickle.load(f)
                    else:
                        # entry = pickle.load(f, encoding='latin1')
                        entry = pickle.load(f)
                    self.data.append(entry['data'])
                    if 'labels' in entry:
                        self.targets.extend(entry['labels'])
                    else:
                        self.targets.extend(entry['fine_labels'])

            self._load_meta()
            self.data = np.vstack(self.data).reshape(-1, im_shape[2], im_shape[0], im_shape[1])
            self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC
        else:
            self.data = np.vstack([p.reshape((1,)+p.shape) for p in data])
            timer.pin_time('reshape')

        print(timer.pin_times_str())

    def array2cifar(self, anArray):
        h, w, ch = anArray.shape
        out = [anArray[:,:,i].reshape(1, h*w) for i in range(ch)]
        return np.hstack(out)



    def _load_meta(self):
        path = os.path.join(self.root, self.base_folder, self.meta['filename'])
        if not check_integrity(path, self.meta['md5']):
            raise RuntimeError('Dataset metadata file not found or corrupted.' +
                               ' You can use download=True to download it')
        with open(path, 'rb') as infile:
            if sys.version_info[0] == 2:
                data = pickle.load(infile)
            else:
                data = pickle.load(infile, encoding='latin1')
            self.classes = data[self.meta['key']]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        # img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)
        if self.indexing: return img, target, index
        else: return img, target

    def __len__(self):
        return len(self.data)

    def _check_integrity(self):
        root = self.root
        for fentry in (self.train_list + self.test_list):
            filename, md5 = fentry[0], fentry[1]
            fpath = os.path.join(root, self.base_folder, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self):
        import tarfile

        if self._check_integrity():
            print('Files already downloaded and verified')
            return

        download_url(self.url, self.root, self.filename, self.tgz_md5)

        # extract file
        with tarfile.open(os.path.join(self.root, self.filename), "r:gz") as tar:
            tar.extractall(path=self.root)

    def extra_repr(self):
        return "Split: {}".format("Train" if self.train is True else "Test")

